#define BG_PROC "WW"
//***************************************************************************
// anl4j.C
//
// JSF macro for analyzing AA -> HH -> 4 jets process at JLC.
//
// (Update Record)
//   2008/12/31 K.Fujii		Original version.
//***************************************************************************

Int_t maxnevt = 20000000;

#ifndef BG_PROC
Char_t *hhspr_so = "../../../higgs/HHStudy/prod/HHSpring.so";
#else
#if BG_PROC == "WW"
Char_t *hhspr_so = "../../../wz/WWStudy/prod/WWSpring.so";
#endif
#endif

#ifndef BG_PROC
int anl4j(const Char_t *infile  = "../../../higgs/HHStudy/prod/data/aa_hhsim.root", 
          const Char_t *outfile = "jsf.hh.root")
#else
#if BG_PROC == "WW"
int anl4j(const Char_t *infile  = "../../../wz/WWStudy/prod/data/aa_wwsim11.gldv4.root",
          const Char_t *outfile = "jsf.ww.11.root")
#endif
#endif
{
  jsf  = new JSFSteer();			// Create JSF object

  gSystem->Load("libS4Utils.so");
  gSystem->Load("libAnlib.so");
  gSystem->Load("libJSFAnlib.so");
  gSystem->Load(hhspr_so);
  gSystem->Load("libHH4JAnalysis.so");

  file = new TFile(outfile,"RECREATE");	// Outputfile
  fin  = new TFile(infile);		// Input simulator data

  jsf->SetInput(*fin);
  jsf->SetOutput(*file);

  // Define modules to use.

  JSFSIMDST     *simdst = new JSFSIMDST();	// Necessary to create SIMDST
  simdst->SetFile(file);			// since we analyze SIMDST
  simdst->NoReadWrite();			// instead of QuickSim data.

  HH4JAnalysis *myanl  = new HH4JAnalysis("HH4JAnalysis","My Analysis");
  Double_t ecm = 379.37; // should be correctly set by hand
  myanl->SetEcm(ecm);
  
  jsf->Initialize();

  JSFQuickSim *sim = (JSFQuickSim*)jsf->FindModule("JSFQuickSim");
  simdst->SetQuickSimParam(sim->Param());

  // Start analysis.

  // Adjust Cut.

  myanl->SetEvisLoCut(0.);	// minimum Evis
  myanl->SetEvisHiCut(420.);	// minimum Evis
  myanl->SetPtCut(999.);	// minimum Pt
  myanl->SetPlCut(999.);	// maximum Pl
  myanl->SetCosjetCut(0.99);	
  myanl->SetMinYcut(0.004);
  myanl->SetM2jLoCut( 90.);
  myanl->SetM2jHiCut(140.);
  myanl->SetMM1Cut( 50.);
  myanl->SetMM2Cut(250.);
#if 1
  // for default detector
  myanl->SetBtagNsig(3.);
  myanl->SetBtagNoffv(1);
  myanl->SetBTtagNsig(3.);
  myanl->SetBTtagNoffv(2);
#else
  // for gldv4 detector
  myanl->SetBtagNsig(2.);
  myanl->SetBtagNoffv(1);
  myanl->SetBTtagNsig(2.);
  myanl->SetBTtagNoffv(2);
#endif
#if 0
  myanl->SetNbjetsCut(4);
  myanl->SetChi2Cut(kTRUE);
#else
  myanl->SetNbjetsCut(0);
  myanl->SetChi2Cut(kFALSE);
#endif

  jsf->BeginRun(1);				// Set run number to 1.
  Int_t nok = 0;
  Int_t nrd = 0;
  while (nok < maxnevt) {
    if (!(jsf->GetEvent(++nrd))) {		// Read in an event.
      JSFSteer::EJSFReturnCode iret = jsf->GetReturnCode();
      if (iret & jsf->kJSFEOF) {
        cerr << " EOF of " << infile << endl;
        cerr << "  reached after reading " << nrd-1 << " events" << endl;
	break;
      } else {
        cerr << " Error reading event " << nrd << " of " << infile << endl;
	cerr << " Terminate analysis" << endl;
	break;
      }
    } else {
        ++nok;
    }
    if (!(jsf->Process(nrd))) continue;	// Do SIMDST and HH4JAnalysis.
  }
  cerr << " Analyzed  " << nok << " events" << endl;
  
  jsf->Terminate();
  return 0;
}
