#define BG_PROC "WW"
//***************************************************************************
// anlhh4j.C
//
// JSF macro for analyzing AA -> HH -> 4 jets process at JLC.
//
// (Update Record)
//   2008/12/31 K.Fujii		Original version.
//***************************************************************************

Int_t maxnevt = 20000000;

Char_t *utils_so = "libS4Utils.so";
Char_t *anlib_so = "libAnlib.so";
Char_t *jsfanlib_so = "libJSFAnlib.so";
Char_t *procanl_so = "libHH4JAnalysis.so";
#ifndef BG_PROC
Char_t *hhspr_so = "../../../higgs/HHStudy/prod/HHSpring.so";
Char_t *outputfile = "jsf.hh.root";	// A file to output histograms
const Int_t kNfiles = 1;
Char_t *inputfile[kNfiles] = {"../../../higgs/HHStudy/prod/data/aa_hhsim.root"};	// Input simulator file
#else
#if BG_PROC == "WW"
Char_t *hhspr_so = "../../../wz/WWStudy/prod/WWSpring.so";
#if 0
Char_t *outputfile = "jsf.ww.root";	// A file to output histograms
const Int_t kNfiles = 6;
Char_t *inputfile[kNfiles] = {"../../../wz/WWStudy/prod/data/aa_wwsim.root",	// Input simulator file
                              "../../../wz/WWStudy/prod/data/aa_wwsim02.root",
                              "../../../wz/WWStudy/prod/data/aa_wwsim03.root",
                              "../../../wz/WWStudy/prod/data/aa_wwsim04.root",
                              "../../../wz/WWStudy/prod/data/aa_wwsim05.root",
                              "../../../wz/WWStudy/prod/data/aa_wwsim06.root"};
#else
#if 1
Char_t *outputfile = "jsf.ww.01.root";	// A file to output histograms
const Int_t kNfiles = 1;
Char_t *inputfile[kNfiles] = {"../../../wz/WWStudy/prod/data/aa_wwsim01.root"};	// Input simulator file
#else
Char_t *outputfile = "jsf.ww.0102.root";	// A file to output histograms
const Int_t kNfiles = 2;
Char_t *inputfile[kNfiles] = {"../../../wz/WWStudy/prod/data/aa_wwsim.root",	// Input simulator file
                              "../../../wz/WWStudy/prod/data/aa_wwsim02.root"};
#endif
#endif
#endif
#endif

int anlhh4j()
{
  jsf  = new JSFSteer();			// Create JSF object

  gSystem->Load(utils_so);
  gSystem->Load(anlib_so);
  gSystem->Load(jsfanlib_so);
  gSystem->Load(hhspr_so);
  gSystem->Load(procanl_so);

  Int_t ifile = 0;
  file = new TFile(outputfile,"RECREATE");	// Outputfile
  fin  = new TFile(inputfile[ifile]);			// Input simulator data

  jsf->SetInput(*fin);
  jsf->SetOutput(*file);

  // Define modules to use.

  JSFSIMDST     *simdst = new JSFSIMDST();	// Necessary to create SIMDST
  simdst->SetFile(file);			// since we analyze SIMDST
  simdst->NoReadWrite();			// instead of QuickSim data.

  HH4JAnalysis *myanl  = new HH4JAnalysis("HH4JAnalysis","My Analysis");
  Double_t ecm = 379.37; // should be correctly set by hand
  myanl->SetEcm(ecm);
  
  jsf->Initialize();

  JSFQuickSim *sim = (JSFQuickSim*)jsf->FindModule("JSFQuickSim");
  simdst->SetQuickSimParam(sim->Param());

  // Start analysis.

  // Adjust Cut.

  myanl->SetEvisLoCut(0.);	// minimum Evis
  myanl->SetEvisHiCut(420.);	// minimum Evis
  myanl->SetPtCut(999.);	// minimum Pt
  myanl->SetPlCut(999.);	// maximum Pl
  myanl->SetCosjetCut(0.99);	
  myanl->SetMinYcut(0.004);
  myanl->SetM2jLoCut( 90.);
  myanl->SetM2jHiCut(140.);
  myanl->SetMM1Cut( 50.);
  myanl->SetMM2Cut(250.);
  myanl->SetBtagNsig(3.);
  myanl->SetBtagNoffv(1);
  myanl->SetBTtagNsig(3.);
  myanl->SetBTtagNoffv(2);
#if 0
  myanl->SetNbjetsCut(4);
  myanl->SetChi2Cut(kTRUE);
#else
  myanl->SetNbjetsCut(0);
  myanl->SetChi2Cut(kFALSE);
#endif

  jsf->BeginRun(1);				// Set run number to 1.
  Int_t nok = 0;
  Int_t nrd = 0;
  while (nok < maxnevt) {
    if (!(jsf->GetEvent(++nrd))) {		// Read in an event.
      JSFSteer::EJSFReturnCode iret = jsf->GetReturnCode();
      if (iret & jsf->kJSFEOF) {
        cerr << " EOF of " << inputfile[ifile] << endl;
        cerr << "  reached after reading " << nrd-1 << " events" << endl;
	ifile++;
        if (ifile >= kNfiles) break;
        fin  = new TFile(inputfile[ifile]);			// Input simulator data
	jsf->SetInput(*fin);
	nrd = 1;
      } else {
        cerr << " Error reading event " << nrd << " of " << inputfile[ifile] << endl;
	cerr << " Terminate analysis" << endl;
	break;
      }
    } else {
        ++nok;
    }
    if (!(jsf->Process(nrd))) continue;	// Do SIMDST and HH4JAnalysis.
  }
  cerr << " Analyzed  " << nok << " events" << endl;
  
  jsf->Terminate();
  return 0;
}
