//****************************************************
//*
//*  Sample UserAnalysis Script
//*
//****************************************************

TCanvas *cHist;
TDirectory *cDir;
TH1F *hNCDC;
TH1F *hNVTX;
TH1F *hNGen;
TH1F *hEHiggs1parton;
TH1F *hEHiggs2parton;
TH1F *hCosHiggs1parton;
TH1F *hCosHiggs2parton;
TH1F *hPabsHiggs1parton;
TH1F *hPabsHiggs2parton;
TNtupleD *hEvt;

//_________________________________________________________
void UserInitialize()
{
  //  This function is called at the begining of the job or when
  //  "reset hist" action is selected in the gui menu.
  //  This is used to define/reset histograms.

  if (! hNCDC) delete hNCDC;
  if (! hNVTX) delete hNVTX;
  if (! hNGen) delete hNGen;
  if (! hEHiggs1parton) delete hEHiggs1parton;
  if (! hEHiggs2parton) delete hEHiggs2parton;
  if (! hCosHiggs1parton) delete hCosHiggs1parton;
  if (! hCosHiggs2parton) delete hCosHiggs2parton;
  if (! hPabsHiggs1parton) delete hPabsHiggs1parton;
  if (! hPabsHiggs2parton) delete hPabsHiggs2parton;
  if (! hEvt) delete hEvt;

  hNCDC = new TH1F("hNCDC", "Number of CDC Tracks",        50, 0, 100);
  hNVTX = new TH1F("hNVTX", "Number of VTX Hits",         100, 0, 200);
  hNGen = new TH1F("hNGen", "Number of Generator Tracks", 150, 0, 300);
  hEHiggs1parton    = new TH1F("hEHiggs1parton",    "Energy of Higgs1 (parton level)",
                               50, 145, 155);
  hEHiggs2parton    = new TH1F("hEHiggs2parton",    "Energy of Higgs2 (parton level)",
                               50, 145, 155);
  hCosHiggs1parton  = new TH1F("hCosHiggs1parton",  "Costh of Higgs1 (parton level)",
                               50, -1, 1);
  hCosHiggs2parton  = new TH1F("hCosHiggs2parton",  "Costh of Higgs2 (parton level)",
                               50, -1, 1);
  hPabsHiggs1parton = new TH1F("hPabsHiggs1parton", "Pabs of Higgs1 (parton level)",
                               50, 0, 100);
  hPabsHiggs2parton = new TH1F("hPabsHiggs2parton", "Pabs of Higgs2 (parton level)",
                               50, 0, 100);
  ntEvt = new TNtupleD("ntEvt", "", "helm:help:erhom:erhop:epim:epip");
  cDir  = gDirectory;
}

//_________________________________________________________
void UserAnalysis()
{
  // This function is called when the processing of one event is completed.
  // Any data processing of the event can be performed in this function.
  //

  cerr << "==== Event No. " << jsf->GetEventNumber() << " ====" << endl;

  JSFSpring    *sp  = static_cast<JSFSpring *>(jsf->FindModule("HHSpring"));
  JSFSpringBuf *spb = static_cast<JSFSpringBuf *>(sp->EventBuf());

  ANL4DVector pcm;
  TClonesArray *partons = spb->GetPartons();

  TIter next(partons);
  JSFSpringParton *parton;
  while ((parton = (JSFSpringParton *)next())) {
#if 0
    cerr << parton->GetSerial() << " "
         << " PID=" << parton->GetID()
         << " Q="   << parton->GetCharge()
         << " M="   << parton->GetMass()
         << " p=("  << parton->GetE() << ","
                    << parton->GetPx() << ","
                    << parton->GetPy() << ","
                    << parton->GetPz() << ")" << endl;
#endif
   if (parton->GetSerial() == 1) {
     hEHiggs1parton->Fill(parton->GetE());
     hCosHiggs1parton->Fill(parton->GetCosth());
     hPabsHiggs1parton->Fill(parton->GetPabs());
   }
   if (parton->GetSerial() == 2) {
     hEHiggs2parton->Fill(parton->GetE());
     hCosHiggs2parton->Fill(parton->GetCosth());
     hPabsHiggs2parton->Fill(parton->GetPabs());
   }
   if (parton->GetNDaughter()) continue;
   pcm += ANL4DVector(parton->GetPV());
  }
#if 0
  cerr << " SpringParton : ";
  pcm.DebugPrint();
#endif
#endif

  // Get event buffer and make combined tracks accessible

//JSFSIMDST    *sds    = static_cast<JSFSIMDST*>(jsf->FindModule("JSFSIMDST"));
  JSFSIMDSTBuf *evt = (JSFSIMDSTBuf *)simdst->EventBuf();

  //  Accumulate information in the histogram

  hNCDC->Fill((Float_t)evt->GetNCDCTracks());
  hNVTX->Fill((Float_t)evt->GetNVTXHits());
  hNGen->Fill((Float_t)evt->GetNGeneratorParticles());

  ANL4DVector qsum;
  TObjArray tracks(1000);
  tracks.SetOwner();
  Int_t    fNtracks = 0;
  Double_t xEtrack = 0.10;

#if 1
  Int_t      ntrks = evt->GetNLTKCLTracks(); // No. of tracks
  TObjArray *trks  = evt->GetLTKCLTracks();  // Combined tracks
  for (Int_t i = 0; i < ntrks; i++) {
    JSFLTKCLTrack *t = static_cast<JSFLTKCLTrack*>(trks->UncheckedAt(i));
    if (t->GetE() > xEtrack) {
      ANLTrack *qt = new ANLTrack(t);
      tracks.Add(qt); // track 4-momentum
      qsum += *qt;    // total 4-momentum
      fNtracks++;     // *qt stays
    }
  }
#else
  Int_t         ntrks = evt->GetNGeneratorParticles();
  TClonesArray *trks   = evt->GetGeneratorParticles();
  for (Int_t i = 0; i < ntrks; i++) {
    JSFGeneratorParticle *t = static_cast<JSFGeneratorParticle *>(trks->UncheckedAt(i));

    Int_t ndau = t->GetNDaughter();
#if 0
    t->ls();
#endif
    if (ndau != 0) continue;
    //printf(" ndau = %d\n", ndau);

    if (t->GetE() > xEtrack) {
      ANL4DVector *qt = new ANL4DVector(t->GetPV());
      tracks.Add(qt);           // track 4-momentum
      qsum += *qt;              // total 4-momentum
      fNtracks++;               // *qt stays
    }
  }
#endif

#if 1
    qsum.DebugPrint();
#endif
}

//_________________________________________________________
void DrawHist()
{
  //  This function is called to draw histograms during the interactive
  //  session.  Thus you can see the accumulation of the histogram
  //  interactively.

  TDirectory *old = gDirectory;
  if (! cHist) {
    cHist = new TCanvas("cHist", "Canvas 1", 0, 0, 800, 800);
    cHist->Divide(3,3);
  }
  else {
    cHist->cd();
  }
  cHist->cd(1);
  hNCDC->Draw();
  cHist->cd(2);
  hNGen->Draw();
  cHist->cd(3);
  hEHiggs1parton->Draw();
  cHist->cd(4);
  hEHiggs2parton->Draw();
  cHist->Update();
  cHist->cd(5);
  hCosHiggs1parton->Draw();
  cHist->cd(6);
  hCosHiggs2parton->Draw();
  cHist->cd(7);
  hPabsHiggs1parton->Draw();
  cHist->cd(8);
  hPabsHiggs2parton->Draw();
  cHist->Update();

  old->cd();
}

//_________________________________________________________
void UserSetOptions()
{
  // This function is called only once, soon after jsf is started.
  // This function can be used to define parameters which is not
  // defined in jsf.conf file.

}

//_________________________________________________________
void UserTerminate()
{
  // This function is called at the end of job.

}
