//*****************************************************************************
//* =====================
//*  ZZSpring
//* =====================
//*
//* (Description)
//*    aa --> ZZ generator
//*
//* (Update Record)
//*    2009/10/21  N.Maeda  Derived from WWSpring.
//*
//*****************************************************************************

#include "JSFSteer.h"
#include "ZZSpring.h"

#include <sstream>
#include <iomanip>
//#define __DEBUG__
//#define __NOLUMI__
//#define __NODECAY__
#ifdef __NODECAY__
#define __ZEROWIDTH__
#endif

//*----------------------------------------------------------------------
//*     Numerical and Natural Constants
//*----------------------------------------------------------------------
#include "GENNumCon.h"

/// Chose Amplitude Table ////////////////////
//#define USE_FT9950
//#define USE_FT199198
//#define USE_FT9951
#define USE_FT100101
//////////////////////////////////////////////
#ifdef USE_FT9950
#include "rFT9950.h"
#include "iFT9950.h"
#endif
#ifdef USE_FT199198
#include "testtableiFT2.h"
#include "testtablerFT2.h"
#endif
#ifdef USE_FT9951
#include "rFT9951.h"
#include "iFT9951.h"
#endif
#ifdef USE_FT100101
#include "rFT100101.h"
#include "iFT100101.h"
#endif

ClassImp(ZZSpring)
ClassImp(ZZSpringBuf)
ClassImp(ZZBases)

extern "C" {
#if 0
extern Complex_t ampzz_(Double_t *gzza, Double_t *mw, Double_t *gamw, Double_t gwf[2],
		    Int_t ihel[2], Double_t kv[2][4],
		    Int_t fhel[4], Double_t mf[4], Double_t pv[4][4]);
#endif
extern void amp_aazz_ft_(Double_t *theta, Double_t *betaz,
			    Int_t fhelinitial[2], Complex_t ft[3][3]);

extern void ephcol_(Int_t *ifl, Double_t *xx, Double_t *ea1, Double_t *ea2,
		    Int_t *ihel1, Int_t *ihel2, Double_t *w);
};

//-----------------------------------------------------------------------------
// ==============================
//  class ZZSpring
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
ZZSpring::ZZSpring(const char *name, const char *title, ZZBases *bases)
          : JSFSpring(name, title, bases)
{
  fEventBuf = new ZZSpringBuf("ZZSpringBuf", "ZZSpring event buffer", this);
  if (!bases) {
    SetBases(new ZZBases);
  }
}

//_____________________________________________________________________________
// --------------------------
//  D-tor
// --------------------------
ZZSpring::~ZZSpring()
{
  //delete fEventBuf;
  //delete GetBases();
}


//-----------------------------------------------------------------------------
// ==============================
//  class ZZSpringBuf
// ==============================
//_____________________________________________________________________________
// --------------------------
//  SetPartons
// --------------------------
Bool_t ZZSpringBuf::SetPartons()
{
  TClonesArray &partons = *fPartons;
  ZZBases      *bases   = static_cast<ZZBases *>(static_cast<ZZSpring *>(Module())->GetBases());

  // ----------------------------------------------
  //  Set 4-momenta of final state particles in CM
  // ----------------------------------------------
  fNparton = 6;

  static ANL4DVector *pv = 0;
  if (!pv) {
    pv = new ANL4DVector [fNparton];
  }
#if 0
  cerr << "---Start SetParton---" << endl;
  cerr << "fP[0](E, Px, Py, Pz) = ("
       << fP[0].E() << ", " << fP[0].Px() << ", "
       << fP[0].Py() << ", " << fP[0].Pz() << ")" << endl;
  cerr << "fP[1](E, Px, Py, Pz) = ("
       << fP[1].E() << ", " << fP[1].Px() << ", "
       << fP[1].Py() << ", " << fP[1].Pz() << ")" << endl;
#endif

#if 0
  cerr << "---Start SetParton---" << endl;
  cerr << "fP[0](E, Px, Py, Pz) = ("
       << (bases->fP[0]).E() << ", " << (bases->fP[0]).Px() << ", " 
       << (bases->fP[0]).Py() << ", " << (bases->fP[0]).Pz() << ")" << endl;
  cerr << "fP[1](E, Px, Py, Pz) = ("
       << (bases->fP[1]).E() << ", " << (bases->fP[1]).Px() << ", " 
       << (bases->fP[1]).Py() << ", " << (bases->fP[1]).Pz() << ")" << endl;
#endif
  pv[2] = bases->fP[0];
  pv[3] = bases->fP[1];
  pv[4] = bases->fP[2];
  pv[5] = bases->fP[3];
  pv[0] = pv[2] + pv[3];
  pv[1] = pv[4] + pv[5];

#if 0
  cerr << "---End SetParton---" << endl;
  cerr << "fP[0](E, Px, Py, Pz) = ("
       << pv[2].E() << ", " << pv[2].Px() << ", " 
       << pv[2].Py() << ", " << pv[2].Pz() << ")" << endl;
  cerr << "fP[1](E, Px, Py, Pz) = ("
       << pv[3].E() << ", " << pv[3].Px() << ", " 
       << pv[3].Py() << ", " << pv[3].Pz() << ")" << endl;
#endif

  // ----------------------------------------------
  //  Boost them to lab. frame
  // ----------------------------------------------

  fZBoost       = bases->GetZBoost();
  fEcmIP        = bases->GetEcmIP();
  fQ2ZZ         = bases->GetQ2ZZ();
  fCosTheta     = bases->GetCosTheta();
  fPhi          = bases->GetPhi();
  fQ2Z1         = bases->GetQ2Z1();
  fCosThetaF1   = bases->GetCosThetaF1();
  fPhiF1        = bases->GetPhiF1();
  fQ2Z2         = bases->GetQ2Z2();
  fCosThetaF2   = bases->GetCosThetaF2();
  fPhiF2        = bases->GetPhiF2();

  Double_t elab = TMath::Sqrt(fEcmIP*fEcmIP + fZBoost*fZBoost);
  TVector3 boostv(0.,0.,fZBoost/elab);

  for (Int_t i=0; i<fNparton; i++) pv[i].Boost(boostv);

  // ----------------------------------------------
  //  Set final state parton infomation
  // ----------------------------------------------

  static TVector **qp = 0;
  if (!qp) {
    qp = new TVector* [fNparton];
    for (Int_t i=0; i<fNparton; i++) qp[i] = new TVector(4);
  }
  for (Int_t i=0; i<fNparton; i++) {
    TVector &q = *qp[i];
    q(0) = pv[i].E ();
    q(1) = pv[i].Px();
    q(2) = pv[i].Py();
    q(3) = pv[i].Pz();
  }
  Int_t    idz    = 23;               // Z PID
  Double_t chz    = 0.;               // Z charge
  Double_t rq2z1  = pv[0].GetMass();
  Double_t rq2z2  = pv[1].GetMass();

  Int_t    id1    = bases->f1Ptr->GetPID();
  Int_t    id2    = bases->f2Ptr->GetPID();
  Double_t m1     = bases->f1Ptr->GetMass();
  Double_t m2     = bases->f2Ptr->GetMass();
  Int_t    hel1   = bases->fHelFinal[0];
  Int_t    hel2   = bases->fHelFinal[1];
  Double_t chg1   = bases->f1Ptr->GetCharge();
  Double_t chg2   = bases->f2Ptr->GetCharge();
  Double_t color1 = bases->f1Ptr->GetColor();
  Int_t    islev1 = color1 > 1. ? 101 : 0;
  Int_t    icf1   = 1;

  Int_t    id3    = bases->f3Ptr->GetPID();
  Int_t    id4    = bases->f4Ptr->GetPID();
  Double_t m3     = bases->f3Ptr->GetMass();
  Double_t m4     = bases->f4Ptr->GetMass();
  Int_t    hel3   = bases->fHelFinal[2];
  Int_t    hel4   = bases->fHelFinal[3];
  Double_t chg3   = bases->f3Ptr->GetCharge();
  Double_t chg4   = bases->f4Ptr->GetCharge();
  Double_t color2 = bases->f3Ptr->GetColor();
  Int_t    islev2 = color2 > 1. ? 201 : 0;
  Int_t    icf2   = 2;

  //                               No.  PID  Mass  Charge    pv  Nd 1st Mom  hel  col shower
  new (partons[0]) JSFSpringParton(  1, idz, rq2z1,  chz, *qp[0], 2,  3,  0,   0,   0,     0);
  new (partons[1]) JSFSpringParton(  2, idz, rq2z2,  chz, *qp[1], 2,  5,  0,   0,   0,     0);
  new (partons[2]) JSFSpringParton(  3, id1,    m1, chg1, *qp[2], 0,  0,  1,hel1,icf1,islev1);
  new (partons[3]) JSFSpringParton(  4, id2,    m2, chg2, *qp[3], 0,  0,  1,hel2,icf1,islev1);
  new (partons[4]) JSFSpringParton(  5, id3,    m3, chg3, *qp[4], 0,  0,  2,hel3,icf2,islev2);
  new (partons[5]) JSFSpringParton(  6, id4,    m4, chg4, *qp[5], 0,  0,  2,hel4,icf2,islev2);

#if 0
  cerr << "------JSFSpringParton PID------" << endl;
  cerr << "PID  Mass  Charge  hel  col shower" << endl;
  cerr << "partons[2] : "
       << id1 << " " << m1 << " " << chg1 << " " << hel1 << " "
       << icf1 << " " << islev1 << endl;
  cerr << "partons[3] : "
       << id2 << " " << m2 << " " << chg2 << " " << hel2 << " "
       << icf1 << " " << islev1 << endl;
  cerr << "partons[4] : "
       << id3 << " " << m3 << " " << chg3 << " " << hel3 << " "
       << icf2 << " " << islev2 << endl;
  cerr << "partons[5] : "
       << id4 << " " << m4 << " " << chg4 << " " << hel4 << " "
       << icf2 << " " << islev2 << endl;
  cerr << endl;
#endif
  return kTRUE ;
}

//-----------------------------------------------------------------------------
// ==============================
//  class ZZBases
// ==============================
//_____________________________________________________________________________
// --------------------------
//  C-tor
// --------------------------
ZZBases::ZZBases(const char *name, const char *title)
         : JSFBases   (name, title),
           fEcm       (300.),
           fZ1ModesLo ( 1),
           fZ1ModesHi (11),
           fZ2ModesLo ( 1),
           fZ2ModesHi (11),
           fZ1BosonPtr( 0),
           fZ2BosonPtr( 0),
           fZBoost    (0.),
           fEcmIP     (fEcm)
{
  //  Constructor of bases.  Default parameter should be initialized here
  //
  // --------------------------------------------
  //  Get parameters from jsf.conf, if specified
  // --------------------------------------------

  cout << "Init zzbases " << endl;

  using namespace std;

  stringstream ins(gJSF->Env()->GetValue("ZZBases.Ecm","300.")); // E_cm
  ins >> fEcm;
  fEcmIP = fEcm;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("ZZBases.Z1ModesLo","1"));  // Z1 decay mode lo
  ins >> fZ1ModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("ZZBases.Z1ModesHi","11")); // Z1 decay mode hi
  ins >> fZ1ModesHi;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("ZZBases.Z2ModesLo","1"));  // Z2 decay mode lo
  ins >> fZ2ModesLo;

  ins.clear();
  ins.str(gJSF->Env()->GetValue("ZZBases.Z2ModesHi","11")); // Z2 decay mode hi
  ins >> fZ2ModesHi;

  // --------------------------------------------
  //  Registor random numbers
  // --------------------------------------------
  //--
  //  Photon initial state helicity
  //--
  DefineVariable(fHelCombFinal, 0., 1., 1, 1);
#ifndef __NOLUMI__
  DefineVariable(fHelAA       , 0., 1., 1, 1);
  DefineVariable(fR_Lum1      , 0., 1., 1, 1);
  DefineVariable(fR_Lum2      , 0., 1., 1, 1);
#else
  DefineVariable(fHelAA       , 0., 1., 0, 1);
  DefineVariable(fR_Lum1      , 0., 1., 0, 1);
  DefineVariable(fR_Lum2      , 0., 1., 0, 1);
#endif

  //--
  //  Z1 and Z2 decay modes
  //--
#ifndef __NODECAY__
  DefineVariable(fZ1DecayMode , 0., 1., 0, 1);
  DefineVariable(fZ2DecayMode , 0., 1., 0, 1);
#else
  DefineVariable(fZ1DecayMode , 0., 1., 1, 1);
  DefineVariable(fZ2DecayMode , 0., 1., 1, 1);
#endif

  //--
  //  Z1 and Z2 invariant masses
  //--
#ifndef __NODECAY__
  DefineVariable(fXQ2Z1       , 0., 1., 1, 1);
  DefineVariable(fXQ2Z2       , 0., 1., 1, 1);
#else
  DefineVariable(fXQ2Z1       , 0., 1., 0, 1);
  DefineVariable(fXQ2Z2       , 0., 1., 0, 1);
#endif

  //--
  //  cos(theta) and phi
  //--
  DefineVariable(fCosTheta  , -1., +1., 1, 1);
  DefineVariable(fPhi       , 0., k2Pi, 0, 1);

#ifndef __NODECAY__
  DefineVariable(fCosThetaF1, -1., +1., 1, 1);
#else 
  DefineVariable(fCosThetaF1, -1., +1., 0, 1);
#endif
  DefineVariable(fPhiF1     , 0., k2Pi, 0, 1);

#ifndef __NODECAY__
  DefineVariable(fCosThetaF2, -1., +1., 1, 1);
#else
  DefineVariable(fCosThetaF2, -1., +1., 0, 1);
#endif
  DefineVariable(fPhiF2     , 0., k2Pi, 0, 1);


  // --------------------------------------------
  //  Set Bases integration parameters
  // --------------------------------------------
  SetNoOfSample(800000);

  SetTuneValue (1.5);
  SetIteration1(0.03, 200);
  SetIteration2(0.02, 500);

}
// --------------------------
//  D-tor
// --------------------------
ZZBases::~ZZBases()
{
  delete fZ1BosonPtr;
  delete fZ2BosonPtr;
}

//_____________________________________________________________________________
// --------------------------
//  Func
// --------------------------
Double_t ZZBases::Func()
{
  //  Bases Integrand
  //

  Double_t bsWeight = 1.;  // Jacobian factor
  Double_t xx[3], ea1, ea2, w;
  Int_t    ifl;
  xx[0] = fR_Lum1;
  xx[1] = fR_Lum2;
  xx[2] = fHelAA;
#if 1
  if (abs(xx[0])>1 || abs(xx[1])>1 || abs(xx[2])>1) {
     cerr << " xx[0]=" << xx[0]
          << " xx[1]=" << xx[1]
          << " xx[2]=" << xx[2]
	  << endl;
  }
#endif
#ifndef __NOLUMI__
  ephcol_(&ifl, xx, &ea1, &ea2, &fHelInitial[0], &fHelInitial[1], &w);
#else
  ea1 = 0.8;
  ea2 = 0.8;
  w   = 1.;
  fHelInitial[0] = -1;
  fHelInitial[1] = -1;
#endif
  ea1 *= fEcm/2;
  ea2 *= fEcm/2;
  fEcmIP   = 2.*TMath::Sqrt(ea1*ea2);     // ignore electron mass 
  bsWeight *= w;

  fZBoost = ea1 - ea2; // P_z of the cm system

  // --------------------------------------------
  //  Select final state
  // --------------------------------------------
  Double_t weight = 1;

#ifndef __NODECAY__
  //fZ1ModePtr = fZ1BosonPtr->PickMode(0.8, weight, fZ1Mode);
  fZ1ModePtr = fZ1BosonPtr->PickMode(fZ1DecayMode, weight, fZ1Mode);
  bsWeight *= weight;
#else
  fZ1ModePtr = fZ1BosonPtr->PickMode(0.3, weight, fZ1Mode);
  if (!fZ1ModePtr) {
     cerr << " fZ1Mode = " << fZ1Mode 
          << " fZ1DecayMode = " << fZ1DecayMode << endl;
  }
#endif
  f1Ptr = static_cast<GENPDTEntry *>(fZ1ModePtr->At(0));
  f2Ptr = static_cast<GENPDTEntry *>(fZ1ModePtr->At(1));
  Double_t m1   = f1Ptr->GetMass();
  Double_t m2   = f2Ptr->GetMass();


#ifndef __NODECAY__
  //fZ2ModePtr = fZ2BosonPtr->PickMode(0.8, weight, fZ2Mode);
  fZ2ModePtr = fZ2BosonPtr->PickMode(fZ2DecayMode, weight, fZ2Mode);
  bsWeight *= weight;
#else
  fZ2ModePtr = fZ2BosonPtr->PickMode(0.3, weight, fZ2Mode);
  if (!fZ2ModePtr) {
     cerr << " fZ2Mode = " << fZ2Mode
          << " fZ2DecayMode = " << fZ2DecayMode << endl;
  }
#endif
  f3Ptr = static_cast<GENPDTEntry *>(fZ2ModePtr->At(0));
  f4Ptr = static_cast<GENPDTEntry *>(fZ2ModePtr->At(1));
  Double_t m3   = f3Ptr->GetMass();
  Double_t m4   = f4Ptr->GetMass();

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod Z-pair.
  // --------------------------------------------
  // temporary treatment
#if 0
  if (fEcmIP < 2*kM_z) {
    return 0.;
  }
#endif

  // --------------------------------------------
  //  Check if reduced Ecm enough for prod fermions.
  // --------------------------------------------
  if (fEcmIP < m1 + m2 + m3 + m4) {
    return 0.;
  }

  // --------------------------------------------
  //  Select helicity combination
  // --------------------------------------------
  //  Notice that spin average for photon is taken
  //  care of here
  SelectHelicities(weight);
  bsWeight *= weight;


  // --------------------------------------------
  //  Decide Q^2 of internal lines
  // --------------------------------------------
  fQ2ZZ = fEcmIP*fEcmIP;
  Double_t rs   = fEcmIP;
#ifndef __ZEROWIDTH__
  Double_t qmin = m1 + m2;
  Double_t qmax = rs - (m3 + m4);
  fQ2Z1 = fZ1BosonPtr->GetQ2BW(qmin, qmax, fXQ2Z1, weight);
  Double_t qz1 = TMath::Sqrt(fQ2Z1);
#else
  fQ2Z1  = TMath::Power(fZ1BosonPtr->GetMass(),2);
  weight = kPi*fZ1BosonPtr->GetMass()*fZ1BosonPtr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  bsWeight *= weight;
#ifndef __ZEROWIDTH__
  qmin  = m3 + m4;
  qmax  = rs - qz1;
  fQ2Z2 = fZ2BosonPtr->GetQ2BW(qmin, qmax, fXQ2Z2, weight);
#else
  fQ2Z2  = TMath::Power(fZ2BosonPtr->GetMass(),2);
  weight = kPi*fZ2BosonPtr->GetMass()*fZ2BosonPtr->GetWidth();
#ifdef __NODECAY__
  weight = 1.;
#endif
#endif
  bsWeight *= weight;

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  GENBranch z1branch(fQ2Z1, fCosThetaF1, fPhiF1, m1*m1    , m2*m2          );
  GENBranch z2branch(fQ2Z2, fCosThetaF2, fPhiF2, m3*m3    , m4*m4          );
  GENBranch cmbranch(fQ2ZZ, fCosTheta  , fPhi  , &z1branch, &z2branch      );

  Double_t sigma = DSigmaDX(cmbranch);

  // --------------------------------------------
  //  Fill plots
  // --------------------------------------------

  H1Fill("h01", fEcmIP           , (bsWeight*sigma));
  H1Fill("h02", fCosTheta        , (bsWeight*sigma));
  H1Fill("h03", fPhi             , (bsWeight*sigma));
  H1Fill("h04", (Double_t)fJCombI, (bsWeight*sigma));

  return (bsWeight * sigma);
}

//_____________________________________________________________________________
// --------------------------
//  DSigmaDX
// --------------------------
Double_t ZZBases::DSigmaDX(GENBranch &cmbranch)
{
  // --------------------------------------------
  //  Phase space
  // --------------------------------------------
  Double_t rs = TMath::Sqrt(cmbranch.GetQ2());
  ANL4DVector qcm(rs, 0., 0., 0.);
  GENFrame    cmframe;
  
  // CM -> Z Z
  Double_t q2z1  = cmbranch.GetM12();
  Double_t q2z2  = cmbranch.GetM22();
  Double_t cosz1 = cmbranch.GetCosTheta();
  Double_t phiz1 = cmbranch.GetPhi     ();
  GENPhase2 phaseCM(qcm, q2z1, q2z2, cmframe, cosz1, phiz1, 0);
  ANL4DVector pz1 = phaseCM.GetFourMomentum(0);
  ANL4DVector pz2 = phaseCM.GetFourMomentum(1);
  Double_t betaz  = phaseCM.GetBetaBar();
  if (betaz <= 0.) return 0.;

  // Z1 -> f1 + f2b
  GENBranch &z1branch = * cmbranch.GetBranchPtr(0);
  Double_t cosf1   = z1branch.GetCosTheta();
  Double_t phif1   = z1branch.GetPhi     ();
  Double_t m12     = z1branch.GetM12();
  Double_t m22     = z1branch.GetM22();

#if 0
  cerr << "----------" << gJSF->GetEventNumber() << "----------" << endl;
  cerr << "betaz = " << betaz << endl;
  cerr << "fP[0] = phaseZ1.GetFourMomentum(0); : "
       << fP[0].E() << " " << fP[0].Px() << " "
       << fP[0].Py() << " " << fP[0].Pz() << endl;
  cerr << "fP[1] = phaseZ1.GetFourMomentum(1); : "
       << fP[1].E() << " " << fP[1].Px() << " "
       << fP[1].Py() << " " << fP[1].Pz() << endl;
#endif 

  GENPhase2 phaseZ1(pz1, m12, m22, cmframe, cosf1, phif1, 1);
  fP[0] = phaseZ1.GetFourMomentum(0);
  fP[1] = phaseZ1.GetFourMomentum(1);
  fM[0] = TMath::Sqrt(m12);
  fM[1] = TMath::Sqrt(m22);
  Double_t betaf1 = phaseZ1.GetBetaBar();
  if (betaf1 <= 0.) return 0.;

#if 0
  cerr << "fP[0] = phaseZ1.GetFourMomentum(0); : "
       << fP[0].E() << " " << fP[0].Px() << " "
       << fP[0].Py() << " " << fP[0].Pz() << endl;
  cerr << "fP[1] = phaseZ1.GetFourMomentum(1); : "
       << fP[1].E() << " " << fP[1].Px() << " "
       << fP[1].Py() << " " << fP[1].Pz() << endl;
  cerr << "betaz = " << betaz << endl;
#endif

  // Z2 -> f3b + f4
  GENBranch &z2branch = * cmbranch.GetBranchPtr(1);
  Double_t cosf2   = z2branch.GetCosTheta();
  Double_t phif2   = z2branch.GetPhi     ();
  Double_t m32     = z2branch.GetM12();
  Double_t m42     = z2branch.GetM22();

  GENPhase2 phaseZ2(pz2, m32, m42, cmframe, cosf2, phif2, 1);
  fP[2] = phaseZ2.GetFourMomentum(0);
  fP[3] = phaseZ2.GetFourMomentum(1);
  fM[2] = TMath::Sqrt(m32);
  fM[3] = TMath::Sqrt(m42);
  Double_t betaf2 = phaseZ2.GetBetaBar();
  if (betaf2 <= 0.) return 0.;

  Double_t eb     = rs/2.;
  Double_t beta_a = 1.;
  fK[0].SetXYZT(0., 0., eb, eb);
  fK[1].SetXYZT(0., 0.,-eb, eb);

  // --------------------------------------------
  //  Calcuate differential cross section
  // --------------------------------------------
  Double_t s      = rs * rs;
#ifdef __DEBUG__
#if 1
  cerr << " fP[0] = (" 
       << fP[0].E () << ","
       << fP[0].Px() << ","
       << fP[0].Py() << ","
       << fP[0].Pz() << ")" << endl;
  cerr << " fP[1] = (" 
       << fP[1].E () << ","
       << fP[1].Px() << ","
       << fP[1].Py() << ","
       << fP[1].Pz() << ")" << endl;
  cerr << " fP[2] = (" 
       << fP[2].E () << ","
       << fP[2].Px() << ","
       << fP[2].Py() << ","
       << fP[2].Pz() << ")" << endl;
  cerr << " fP[3] = (" 
       << fP[3].E () << ","
       << fP[3].Px() << ","
       << fP[3].Py() << ","
       << fP[3].Pz() << ")" << endl;
#endif
  ANL4DVector qz1 = fP[0] + fP[1];
  ANL4DVector qz2 = fP[2] + fP[3];
  cerr << " qz1 = (" 
       << qz1.E () << ","
       << qz1.Px() << ","
       << qz1.Py() << ","
       << qz1.Pz() << ")" << endl;
  cerr << " qz2 = (" 
       << qz2.E () << ","
       << qz2.Px() << ","
       << qz2.Py() << ","
       << qz2.Pz() << ")" << endl;
  cerr << " mz1 = " << qz1.GetMass() << " mz2 = " << qz2.GetMass() << endl;
  ANL4DVector pcm = qz1 + qz2;
  cerr << " pcm = (" 
       << pcm.E () << ","
       << pcm.Px() << ","
       << pcm.Py() << ","
       << pcm.Pz() << ")" << endl;
#endif

  // -------------------
  //  Amplitude squared
  // -------------------
  //  Complex_t amp  = AmpAAtoZZ();
#if 1
#if 0
  cerr << "before AmpAAtoZZ " << endl;
  cerr << "fP[0] = phaseZ1.GetFourMomentum(0); : "
       << fP[0].E() << " " << fP[0].Px() << " "
       << fP[0].Py() << " " << fP[0].Pz() << endl;
  cerr << "fP[1] = phaseZ1.GetFourMomentum(1); : "
       << fP[1].E() << " " << fP[1].Px() << " "
       << fP[1].Py() << " " << fP[1].Pz() << endl;
  cerr << "betaz = " << betaz << endl;
#endif
  Complex_t amp  = AmpAAtoZZ(betaz);
  Double_t  amp2 = (amp*conj(amp)).real();
#if 0
  cerr << "end AmpAAtoZZ " << endl;
  cerr << "fP[0] = phaseZ1.GetFourMomentum(0); : "
       << fP[0].E() << " " << fP[0].Px() << " "
       << fP[0].Py() << " " << fP[0].Pz() << endl;
  cerr << "fP[1] = phaseZ1.GetFourMomentum(1); : "
       << fP[1].E() << " " << fP[1].Px() << " "
       << fP[1].Py() << " " << fP[1].Pz() << endl;
  cerr << "betaz = " << betaz << endl;
#endif
#else
  Complex_t amp  = 1;
  Double_t  amp2 = 1;
#endif

  // -------------------
  //  Put them together
  // -------------------
#ifndef __NODECAY__
  static const Int_t    kNbr  = 3;
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3*kNbr));
#else
  static const Double_t kFact = k2Pi/(TMath::Power(k4Pi,3+2));
  betaf1 = 1.;
  betaf2 = 1.;
#endif

#if 1
  Double_t identp = 0.5;                           // identical particle factor
  Double_t dPhase = kFact * betaz * betaf1*betaf2; // phase space factor
  Double_t flux   = 1./(2.* s * beta_a);           // beam flux factor

  Double_t sigma  = identp * flux * amp2 * dPhase; // in [1/GeV^2]
           sigma *= kGeV2fb;                       // now in [fb]
#else	   
  Double_t sigma = 1/(128.*TMath::Pi()) * amp2 * betaz / s;
           sigma *= kGeV2fb;
#endif


  return sigma;

}

// --------------------------
//  AmpAAtoZZ()
// --------------------------
Complex_t ZZBases::AmpAAtoZZ(Double_t betaz)
{
#if 1

#if 0
  cerr << "fP[0]:" << fP[0].E() << endl;
  cerr << "fP[1]:" << fP[1].E() << endl;
  cerr << "fP[2]:" << fP[2].E() << endl;
  cerr << "fP[3]:" << fP[3].E() << endl;
#endif

#if 0
  Double_t mz1   = fZ1BosonPtr->GetMass ();
  Double_t gamz1 = fZ1BosonPtr->GetWidth();
  Double_t mz2   = fZ2BosonPtr->GetMass ();
  Double_t gamz2 = fZ2BosonPtr->GetWidth();
  Double_t gzza  = kGe;
  Double_t gzf[2];
  gzf[0] = -kGz*kSqh;
  gzf[1] = 0.;

  ANL4DVector fPZ[2];  // [0, 1] = [z1, z2]
  fPZ[0] = fP[0] + fP[1];
  fPZ[1] = fP[2] + fP[3];
  
  Double_t s  = (fK[0] + fK[1]).GetMass2();
  Double_t t  = (fPZ[0] - fK[0]).GetMass2();
  Double_t u  = (fPZ[1] - fK[0]).GetMass2();

  static const Double_t al_em = 1./128.;
  static const Double_t al_w  = 1./30.;
#endif

#if 0
  if (fHelInitial[0] == fHelInitial[1]) {
    amp = amp_pp_(&s, &t, &u, &mh, &mt, &mw, &fDK);
  } else {
    amp = amp_pm_(&s, &t, &u, &mh, &mt, &mw, &fDK);
  }
  amp *= al_em * al_w;
#endif

  // temporaliry value used
#if 1
  ANL4DVector fPZ[2];  // [0, 1] = [z1, z2]
  fPZ[0] = fP[0] + fP[1];
  fPZ[1] = fP[2] + fP[3];
  Double_t theta = TMath::Pi()*(fPZ[0].GetTheta())/180.;
#else
  // for Debugging
  Double_t theta = TMath::Pi();
#endif


#if 0
  cerr << "theta = " << theta << endl;
  cerr << "betaz = " << betaz << endl;
  cerr << "bz = " << bz << endl;
  cerr << "fHelInitial = " << fHelInitial[0] << " " << fHelInitial[1] << endl;
  //  cerr << "s = " << s << endl;
  cerr << "kM_z = " << kM_z << endl;
  cerr << "photon1 energy = " << fK[0].E() << endl;
  cerr << "photon2 energy = " << fK[1].E() << endl;
#endif 
  Complex_t FT[3][3];

  Double_t cutTheta_Lo, cutTheta_Hi;
#if defined(USE_FT9950) || defined(USE_FT199198)
  cutTheta_Lo = 8.*TMath::Pi()/180.;
  cutTheta_Hi = 172.*TMath::Pi()/180.;
#endif
#if defined(USE_FT9951) || defined(USE_FT100101)
  cutTheta_Lo = 0.5*TMath::Pi()/180.;
  cutTheta_Hi = 179.5*TMath::Pi()/180.;
#endif

  Double_t cutBetaz;
#if defined(USE_FT9950) || defined(USE_FT9951)
  cutBetaz = 0.01;
#endif
#if defined(USE_FT199198) || defined(USE_FT100101) 
  cutBetaz = 0.005;
#endif
#if 0
  if(betaz < cutBetaz) {
    cerr << "betaz = " << betaz << endl;
  }
#endif

  if(cutTheta_Lo<theta && theta<cutTheta_Hi && betaz >= cutBetaz){
    //amp_aazz_ft_(&theta, &betaz, fHelInitial, FT);
    TakeAmp(fHelInitial, theta, betaz, FT);
  } else {
    for(int i=0;i<3;i++){
      for(int j=0;j<3;j++){
	FT[i][j] = Complex_t(0.,0.);
	//cerr << "FT" << i << j << " = " << FT[i][j] << endl;
      }
    }
  }

#ifndef __NODECAY__
  static const Bool_t kIsIncoming = kTRUE;
  static const Bool_t kIsOutgoing = kFALSE;

  Double_t gamz   = fZ1BosonPtr->GetWidth();

  Double_t qf1    = f1Ptr->GetCharge();
  Double_t t3f1   = f1Ptr->GetISpin();
  Double_t glz1   = -kGz*(t3f1 - qf1*kSin2W);
  Double_t grz1   = -kGz*(     - qf1*kSin2W);

  HELFermion f1  (fP[0], fM[0], fHelFinal[0], +1, kIsOutgoing);
  HELFermion f2b (fP[1], fM[1], fHelFinal[1], -1, kIsIncoming);
  HELVector  z1f (f2b, f1, glz1, grz1, kM_z, gamz);
  
  // temporarily on-shell mass used
  HELVector  z1im(fP[0]+fP[1], kM_z, -1, -1);
  HELVector  z1i0(fP[0]+fP[1], kM_z,  0, -1);
  HELVector  z1ip(fP[0]+fP[1], kM_z, +1, -1);
 
  Complex_t z1mAmp = z1f[0]*z1im[0] - z1f[1]*z1im[1] - z1f[2]*z1im[2] - z1f[3]*z1im[3];
  Complex_t z10Amp = z1f[0]*z1i0[0] - z1f[1]*z1i0[1] - z1f[2]*z1i0[2] - z1f[3]*z1i0[3];
  Complex_t z1pAmp = z1f[0]*z1ip[0] - z1f[1]*z1ip[1] - z1f[2]*z1ip[2] - z1f[3]*z1ip[3];

  Double_t qf3    = f3Ptr->GetCharge();
  Double_t t3f3   = f3Ptr->GetISpin();
  Double_t glz2   = -kGz*(t3f3 - qf3*kSin2W);
  Double_t grz2   = -kGz*(     - qf3*kSin2W);

  HELFermion f3b (fP[2], fM[2], fHelFinal[2], -1, kIsIncoming);
  HELFermion f4  (fP[3], fM[3], fHelFinal[3], +1, kIsOutgoing);
  HELVector  z2f (f3b, f4, glz2, grz2, kM_z, gamz);
 
  // temporarily on-shell mass used
  HELVector  z2im(fP[2]+fP[3], kM_z, -1, -1);
  HELVector  z2i0(fP[2]+fP[3], kM_z,  0, -1);
  HELVector  z2ip(fP[2]+fP[3], kM_z, +1, -1);

  Complex_t z2mAmp = z2f[0]*z2im[0] - z2f[1]*z2im[1] - z2f[2]*z2im[2] - z2f[3]*z2im[3];
  Complex_t z20Amp = z2f[0]*z2i0[0] - z2f[1]*z2i0[1] - z2f[2]*z2i0[2] - z2f[3]*z2i0[3];
  Complex_t z2pAmp = z2f[0]*z2ip[0] - z2f[1]*z2ip[1] - z2f[2]*z2ip[2] - z2f[3]*z2ip[3];

  Complex_t amp;
  amp =
    FT[0][0]*z1mAmp*z2mAmp +
    FT[0][1]*z1mAmp*z20Amp +
    FT[0][2]*z1mAmp*z2pAmp +
    FT[1][0]*z10Amp*z2mAmp +
    FT[1][1]*z10Amp*z20Amp +
    FT[1][2]*z10Amp*z2pAmp +
    FT[2][0]*z1pAmp*z2mAmp +
    FT[2][1]*z1pAmp*z20Amp +
    FT[2][2]*z1pAmp*z2pAmp;

  Double_t color = TMath::Sqrt(f1Ptr->GetColor() * f3Ptr->GetColor());

#else
  Int_t i = 0;
  Int_t j = 0;
  if (fHelFinal[0] == -1) {
    i = 0;
  } else if (fHelFinal[0] == 0) {
    i = 1;
  } else if (fHelFinal[0] == +1) {
    i = 2;
  }

  if (fHelFinal[1] == -1) {
    j = 0;
  } else if (fHelFinal[1] == 0) {
    j = 1;
  } else if (fHelFinal[1] == +1) {
    j = 2;
  }

  Complex_t amp;
  amp = FT[i][j];
  Double_t color = 1;
#endif

#if 0
  cerr << "Initial helicity = "
       << fHelInitial[0] << ", " << fHelInitial[1] << endl;
  cerr << "Final helicity = "
       << fHelFinal[0] << ", " << fHelFinal[1] << endl;
  cerr << "betaz = " << betaz << endl;
  cerr << "theta = " << theta << endl;
  cerr << "amplitude = " << amp << endl;
  cerr << endl;
#endif

  amp *= color;
  return amp;
#else 

  return Complex_t(0, 0);
#endif
}

//_____________________________________________________________________________
// --------------------------
//  Userin
// --------------------------
void ZZBases::Userin()
{
  // --------------------------------------------
  //  Initialize decay table
  // --------------------------------------------
  if (!fZ1BosonPtr) fZ1BosonPtr = new GENPDTZBoson();
  for (Int_t m=1; m<=fZ1BosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fZ1BosonPtr->GetMode(m); 
     if (mp && (m<fZ1ModesLo || m>fZ1ModesHi)) {
        mp->Lock();
     }
  }
  fZ1BosonPtr->DebugPrint();
  if (!fZ2BosonPtr) fZ2BosonPtr = new GENPDTZBoson();
  for (Int_t m=1; m<=fZ2BosonPtr->GetEntries(); m++) {
     GENDecayMode *mp = fZ2BosonPtr->GetMode(m); 
     if (mp && (m<fZ2ModesLo || m>fZ2ModesHi)) {
        mp->Lock();
     }
  }
  fZ2BosonPtr->DebugPrint();

  // --------------------------------------------
  //  Define some plots
  // --------------------------------------------
  H1Init("h01", "Ecm"   , 100,     0., fEcm*1.1);
  H1Init("h02", "Costh" , 100,    -1.,      +1.);
  H1Init("h03", "Phi"   , 100,     0.,     k2Pi);
  H1Init("h04", "Helin" ,   4,     0.,       4.);
}

//_____________________________________________________________________________
// --------------------------
//  Userout
// --------------------------
void ZZBases::Userout()
{
  cout << "End of ZZBases----------------------------------- "  << endl
       << "Ecm                  = " << fEcm  << " [GeV]   "     << endl
       << "Total Cross section  = " << GetEstimate()  << " +/- "
                                    << GetError()     << " [fb]"<< endl
       << "Number of iterations = " << GetNoOfIterate()         << endl;
}

//_____________________________________________________________________________
// --------------------------
//  SelectHelicities
// --------------------------
void ZZBases::SelectHelicities(Double_t &weight)
{
#ifndef __NODECAY__
   static const Int_t kNf = 4;
   static const Int_t kFHelComb[kNf][4] = {{+1, -1, -1, +1},
					   {+1, -1, +1, -1},
					   {-1, +1, -1, +1},
					   {-1, +1, +1, -1}};
#else
   static const Int_t kNf = 9;
   static const Int_t kFHelComb[kNf][4] = {{-1, -1, -1, +1},
                                           {-1,  0, -1, +1},
                                           {-1, +1, -1, +1},
                                           { 0, -1, -1, +1},
                                           { 0,  0, -1, +1},
                                           { 0, +1, -1, +1},
                                           {+1, -1, -1, +1},
                                           {+1,  0, -1, +1},
                                           {+1, +1, -1, +1}};
#endif
   fJCombI =  2*( (fHelInitial[0] + 1)/2 ) + (fHelInitial[1] + 1)/2;

   fJCombF = (Int_t)(fHelCombFinal*kNf);
   fJCombF = TMath::Min(fJCombF, kNf-1);
   fHelFinal  [0] = kFHelComb[fJCombF][0];
   fHelFinal  [1] = kFHelComb[fJCombF][1];
   fHelFinal  [2] = kFHelComb[fJCombF][2];
   fHelFinal  [3] = kFHelComb[fJCombF][3];
   weight = kNf;
}
//__________________________________________________________
// --------------------------
//  take amplitude from realFT.h and imagFT.h
// --------------------------
void ZZBases::TakeAmp(Int_t fHelInitial[2], Double_t theta, Double_t betaz,
                      Complex_t FT[3][3]){
  Int_t n, m;
#ifdef USE_FT9950
  Double_t deltat = 3.3469388*TMath::Pi()/180.;
  Double_t deltab = 0.01;
  Double_t initialt = 8.*TMath::Pi()/180.;
  Double_t initialb = 0.01;     //initial betaz value in table.
  const Int_t kNmax = 99;
  const Int_t kMmax = 50;
#endif
#ifdef USE_FT199198
  Double_t deltat = 3.3469388/4*TMath::Pi()/180.;
  Double_t deltab = 0.005;
  Double_t initialt = 8.*TMath::Pi()/180.;
  Double_t initialb = 0.005;     //initial betaz value in table.
  const Int_t kNmax = 199;
  const Int_t kMmax = 198;
#endif
#ifdef USE_FT9951
  Double_t deltat = 3.58*TMath::Pi()/180.;
  Double_t deltab = 0.01;
  Double_t initialt = 0.5*TMath::Pi()/180.;
  Double_t initialb = 0.015;     //initial betaz value in table.
  const Int_t kNmax = 99;
  const Int_t kMmax = 51;
#endif
#ifdef USE_FT100101
  Double_t deltat = 1.79*TMath::Pi()/180.;
  Double_t deltab = 0.01;
  Double_t initialt = 0.5*TMath::Pi()/180.;
  Double_t initialb = 0.005;     //initial betaz value in table.
  const Int_t kNmax = 100;
  const Int_t kMmax = 101;
#endif

  Double_t fractiont, fractionb;
  n = (betaz-initialb) / deltab;
  m = (theta-initialt) / deltat;

  if (n < 0) n = 0;
  if (n >= kNmax-2) n = kNmax-2;
  if (m < 0) m = 0;
  if (m >= kMmax-2) m = kMmax-2;

  fractiont = (theta - (initialt + deltat*m));
  fractionb = (betaz - (initialb + deltab*n));
#if 0
  if (fractiont < 0) fractiont = 0;
  if (fractiont > deltat) fractiont = deltat;
#endif
  //  cerr << "betaz = " << betaz << "  theta = " << theta << endl;
  //  cerr << "n = " << n << "  m = " << m << endl;
  Double_t rFT[3][3];
  Double_t iFT[3][3];
  Double_t rFTmb, rFTm1b;
  Double_t iFTmb, iFTm1b;

  Int_t initial = 0;
  if(fHelInitial[0]==1 && fHelInitial[1]==1){
    initial = 1;
  } else if(fHelInitial[0]==-1 && fHelInitial[1]==-1){
    initial = 2;
  } else if(fHelInitial[0]==1 && fHelInitial[1]==-1){
    initial = 3;
  } else if(fHelInitial[0]==-1 && fHelInitial[1]==1){
    initial = 4;
  } else {
    return;
  }

  if(initial==1) {
    rFTmb = krFTpppp[n][m]
      + (fractionb * (krFTpppp[n+1][m]-krFTpppp[n][m])/deltab);
    rFTm1b = krFTpppp[n][m+1]
      + (fractionb * (krFTpppp[n+1][m+1]-krFTpppp[n][m+1])/deltab);
    rFT[2][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpppp[n][m]
      + (fractionb * (kiFTpppp[n+1][m]-kiFTpppp[n][m])/deltab);
    iFTm1b = kiFTpppp[n][m+1]
      + (fractionb * (kiFTpppp[n+1][m+1]-kiFTpppp[n][m+1])/deltab);
    iFT[2][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);
    
    rFTmb = krFTppp0[n][m]
      + (fractionb * (krFTppp0[n+1][m]-krFTppp0[n][m])/deltab);
    rFTm1b = krFTppp0[n][m+1]
      + (fractionb * (krFTppp0[n+1][m+1]-krFTppp0[n][m+1])/deltab);
    rFT[2][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTppp0[n][m]
      + (fractionb * (kiFTppp0[n+1][m]-kiFTppp0[n][m])/deltab);
    iFTm1b = kiFTppp0[n][m+1]
      + (fractionb * (kiFTppp0[n+1][m+1]-kiFTppp0[n][m+1])/deltab);
    iFT[2][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpppm[n][m]
      + (fractionb * (krFTpppm[n+1][m]-krFTpppm[n][m])/deltab);
    rFTm1b = krFTpppm[n][m+1]
      + (fractionb * (krFTpppm[n+1][m+1]-krFTpppm[n][m+1])/deltab);
    rFT[2][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpppm[n][m]
      + (fractionb * (kiFTpppm[n+1][m]-kiFTpppm[n][m])/deltab);
    iFTm1b = kiFTpppm[n][m+1]
      + (fractionb * (kiFTpppm[n+1][m+1]-kiFTpppm[n][m+1])/deltab);
    iFT[2][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);


    rFTmb = krFTpp0p[n][m]
      + (fractionb * (krFTpp0p[n+1][m]-krFTpp0p[n][m])/deltab);
    rFTm1b = krFTpp0p[n][m+1]
      + (fractionb * (krFTpp0p[n+1][m+1]-krFTpp0p[n][m+1])/deltab);
    rFT[1][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpp0p[n][m]
      + (fractionb * (kiFTpp0p[n+1][m]-kiFTpp0p[n][m])/deltab);
    iFTm1b = kiFTpp0p[n][m+1]
      + (fractionb * (kiFTpp0p[n+1][m+1]-kiFTpp0p[n][m+1])/deltab);
    iFT[1][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpp00[n][m]
      + (fractionb * (krFTpp00[n+1][m]-krFTpp00[n][m])/deltab);
    rFTm1b = krFTpp00[n][m+1]
      + (fractionb * (krFTpp00[n+1][m+1]-krFTpp00[n][m+1])/deltab);
    rFT[1][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpp00[n][m]
      + (fractionb * (kiFTpp00[n+1][m]-kiFTpp00[n][m])/deltab);
    iFTm1b = kiFTpp00[n][m+1]
      + (fractionb * (kiFTpp00[n+1][m+1]-kiFTpp00[n][m+1])/deltab);
    iFT[1][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpp0m[n][m]
      + (fractionb * (krFTpp0m[n+1][m]-krFTpp0m[n][m])/deltab);
    rFTm1b = krFTpp0m[n][m+1]
      + (fractionb * (krFTpp0m[n+1][m+1]-krFTpp0m[n][m+1])/deltab);
    rFT[1][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpp0m[n][m]
      + (fractionb * (kiFTpp0m[n+1][m]-kiFTpp0m[n][m])/deltab);
    iFTm1b = kiFTpp0m[n][m+1]
      + (fractionb * (kiFTpp0m[n+1][m+1]-kiFTpp0m[n][m+1])/deltab);
    iFT[1][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTppmp[n][m]
      + (fractionb * (krFTppmp[n+1][m]-krFTppmp[n][m])/deltab);
    rFTm1b = krFTppmp[n][m+1]
      + (fractionb * (krFTppmp[n+1][m+1]-krFTppmp[n][m+1])/deltab);
    rFT[0][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTppmp[n][m]
      + (fractionb * (kiFTppmp[n+1][m]-kiFTppmp[n][m])/deltab);
    iFTm1b = kiFTppmp[n][m+1]
      + (fractionb * (kiFTppmp[n+1][m+1]-kiFTppmp[n][m+1])/deltab);
    iFT[0][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTppm0[n][m]
      + (fractionb * (krFTppm0[n+1][m]-krFTppm0[n][m])/deltab);
    rFTm1b = krFTppm0[n][m+1]
      + (fractionb * (krFTppm0[n+1][m+1]-krFTppm0[n][m+1])/deltab);
    rFT[0][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTppm0[n][m]
      + (fractionb * (kiFTppm0[n+1][m]-kiFTppm0[n][m])/deltab);
    iFTm1b = kiFTppm0[n][m+1]
      + (fractionb * (kiFTppm0[n+1][m+1]-kiFTppm0[n][m+1])/deltab);
    iFT[0][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTppmm[n][m]
      + (fractionb * (krFTppmm[n+1][m]-krFTppmm[n][m])/deltab);
    rFTm1b = krFTppmm[n][m+1]
      + (fractionb * (krFTppmm[n+1][m+1]-krFTppmm[n][m+1])/deltab);
    rFT[0][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTppmm[n][m]
      + (fractionb * (kiFTppmm[n+1][m]-kiFTppmm[n][m])/deltab);
    iFTm1b = kiFTppmm[n][m+1]
      + (fractionb * (kiFTppmm[n+1][m+1]-kiFTppmm[n][m+1])/deltab);
    iFT[0][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);
  } else if(initial==2) {
    rFTmb = krFTmmpp[n][m]
      + (fractionb * (krFTmmpp[n+1][m]-krFTmmpp[n][m])/deltab);
    rFTm1b = krFTmmpp[n][m+1]
      + (fractionb * (krFTmmpp[n+1][m+1]-krFTmmpp[n][m+1])/deltab);
    rFT[2][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmmpp[n][m]
      + (fractionb * (kiFTmmpp[n+1][m]-kiFTmmpp[n][m])/deltab);
    iFTm1b = kiFTmmpp[n][m+1]
      + (fractionb * (kiFTmmpp[n+1][m+1]-kiFTmmpp[n][m+1])/deltab);
    iFT[2][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmmp0[n][m]
      + (fractionb * (krFTmmp0[n+1][m]-krFTmmp0[n][m])/deltab);
    rFTm1b = krFTmmp0[n][m+1]
      + (fractionb * (krFTmmp0[n+1][m+1]-krFTmmp0[n][m+1])/deltab);
    rFT[2][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmmp0[n][m]
      + (fractionb * (kiFTmmp0[n+1][m]-kiFTmmp0[n][m])/deltab);
    iFTm1b = kiFTmmp0[n][m+1]
      + (fractionb * (kiFTmmp0[n+1][m+1]-kiFTmmp0[n][m+1])/deltab);
    iFT[2][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmmpm[n][m]
      + (fractionb * (krFTmmpm[n+1][m]-krFTmmpm[n][m])/deltab);
    rFTm1b = krFTmmpm[n][m+1]
      + (fractionb * (krFTmmpm[n+1][m+1]-krFTmmpm[n][m+1])/deltab);
    rFT[2][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmmpm[n][m]
      + (fractionb * (kiFTmmpm[n+1][m]-kiFTmmpm[n][m])/deltab);
    iFTm1b = kiFTmmpm[n][m+1]
      + (fractionb * (kiFTmmpm[n+1][m+1]-kiFTmmpm[n][m+1])/deltab);
    iFT[2][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmm0p[n][m]
      + (fractionb * (krFTmm0p[n+1][m]-krFTmm0p[n][m])/deltab);
    rFTm1b = krFTmm0p[n][m+1]
      + (fractionb * (krFTmm0p[n+1][m+1]-krFTmm0p[n][m+1])/deltab);
    rFT[1][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmm0p[n][m]
      + (fractionb * (kiFTmm0p[n+1][m]-kiFTmm0p[n][m])/deltab);
    iFTm1b = kiFTmm0p[n][m+1]
      + (fractionb * (kiFTmm0p[n+1][m+1]-kiFTmm0p[n][m+1])/deltab);
    iFT[1][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmm00[n][m]
      + (fractionb * (krFTmm00[n+1][m]-krFTmm00[n][m])/deltab);
    rFTm1b = krFTmm00[n][m+1]
      + (fractionb * (krFTmm00[n+1][m+1]-krFTmm00[n][m+1])/deltab);
    rFT[1][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmm00[n][m]
      + (fractionb * (kiFTmm00[n+1][m]-kiFTmm00[n][m])/deltab);
    iFTm1b = kiFTmm00[n][m+1]
      + (fractionb * (kiFTmm00[n+1][m+1]-kiFTmm00[n][m+1])/deltab);
    iFT[1][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmm0m[n][m]
      + (fractionb * (krFTmm0m[n+1][m]-krFTmm0m[n][m])/deltab);
    rFTm1b = krFTmm0m[n][m+1]
      + (fractionb * (krFTmm0m[n+1][m+1]-krFTmm0m[n][m+1])/deltab);
    rFT[1][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmm0m[n][m]
      + (fractionb * (kiFTmm0m[n+1][m]-kiFTmm0m[n][m])/deltab);
    iFTm1b = kiFTmm0m[n][m+1]
      + (fractionb * (kiFTmm0m[n+1][m+1]-kiFTmm0m[n][m+1])/deltab);
    iFT[1][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmmmp[n][m]
      + (fractionb * (krFTmmmp[n+1][m]-krFTmmmp[n][m])/deltab);
    rFTm1b = krFTmmmp[n][m+1]
      + (fractionb * (krFTmmmp[n+1][m+1]-krFTmmmp[n][m+1])/deltab);
    rFT[0][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmmmp[n][m]
      + (fractionb * (kiFTmmmp[n+1][m]-kiFTmmmp[n][m])/deltab);
    iFTm1b = kiFTmmmp[n][m+1]
      + (fractionb * (kiFTmmmp[n+1][m+1]-kiFTmmmp[n][m+1])/deltab);
    iFT[0][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmmm0[n][m]
      + (fractionb * (krFTmmm0[n+1][m]-krFTmmm0[n][m])/deltab);
    rFTm1b = krFTmmm0[n][m+1]
      + (fractionb * (krFTmmm0[n+1][m+1]-krFTmmm0[n][m+1])/deltab);
    rFT[0][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmmm0[n][m]
      + (fractionb * (kiFTmmm0[n+1][m]-kiFTmmm0[n][m])/deltab);
    iFTm1b = kiFTmmm0[n][m+1]
      + (fractionb * (kiFTmmm0[n+1][m+1]-kiFTmmm0[n][m+1])/deltab);
    iFT[0][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmmmm[n][m]
      + (fractionb * (krFTmmmm[n+1][m]-krFTmmmm[n][m])/deltab);
    rFTm1b = krFTmmmm[n][m+1]
      + (fractionb * (krFTmmmm[n+1][m+1]-krFTmmmm[n][m+1])/deltab);
    rFT[0][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmmmm[n][m]
      + (fractionb * (kiFTmmmm[n+1][m]-kiFTmmmm[n][m])/deltab);
    iFTm1b = kiFTmmmm[n][m+1]
      + (fractionb * (kiFTmmmm[n+1][m+1]-kiFTmmmm[n][m+1])/deltab);
    iFT[0][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);
  } else if(initial==3) {
    rFTmb = krFTpmpp[n][m]
      + (fractionb * (krFTpmpp[n+1][m]-krFTpmpp[n][m])/deltab);
    rFTm1b = krFTpmpp[n][m+1]
      + (fractionb * (krFTpmpp[n+1][m+1]-krFTpmpp[n][m+1])/deltab);
    rFT[2][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpmpp[n][m]
      + (fractionb * (kiFTpmpp[n+1][m]-kiFTpmpp[n][m])/deltab);
    iFTm1b = kiFTpmpp[n][m+1]
      + (fractionb * (kiFTpmpp[n+1][m+1]-kiFTpmpp[n][m+1])/deltab);
    iFT[2][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpmp0[n][m]
      + (fractionb * (krFTpmp0[n+1][m]-krFTpmp0[n][m])/deltab);
    rFTm1b = krFTpmp0[n][m+1]
      + (fractionb * (krFTpmp0[n+1][m+1]-krFTpmp0[n][m+1])/deltab);
    rFT[2][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpmp0[n][m]
      + (fractionb * (kiFTpmp0[n+1][m]-kiFTpmp0[n][m])/deltab);
    iFTm1b = kiFTpmp0[n][m+1]
      + (fractionb * (kiFTpmp0[n+1][m+1]-kiFTpmp0[n][m+1])/deltab);
    iFT[2][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpmpm[n][m]
      + (fractionb * (krFTpmpm[n+1][m]-krFTpmpm[n][m])/deltab);
    rFTm1b = krFTpmpm[n][m+1]
      + (fractionb * (krFTpmpm[n+1][m+1]-krFTpmpm[n][m+1])/deltab);
    rFT[2][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpmpm[n][m]
      + (fractionb * (kiFTpmpm[n+1][m]-kiFTpmpm[n][m])/deltab);
    iFTm1b = kiFTpmpm[n][m+1]
      + (fractionb * (kiFTpmpm[n+1][m+1]-kiFTpmpm[n][m+1])/deltab);
    iFT[2][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpm0p[n][m]
      + (fractionb * (krFTpm0p[n+1][m]-krFTpm0p[n][m])/deltab);
    rFTm1b = krFTpm0p[n][m+1]
      + (fractionb * (krFTpm0p[n+1][m+1]-krFTpm0p[n][m+1])/deltab);
    rFT[1][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpm0p[n][m]
      + (fractionb * (kiFTpm0p[n+1][m]-kiFTpm0p[n][m])/deltab);
    iFTm1b = kiFTpm0p[n][m+1]
      + (fractionb * (kiFTpm0p[n+1][m+1]-kiFTpm0p[n][m+1])/deltab);
    iFT[1][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpm00[n][m]
      + (fractionb * (krFTpm00[n+1][m]-krFTpm00[n][m])/deltab);
    rFTm1b = krFTpm00[n][m+1]
      + (fractionb * (krFTpm00[n+1][m+1]-krFTpm00[n][m+1])/deltab);
    rFT[1][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpm00[n][m]
      + (fractionb * (kiFTpm00[n+1][m]-kiFTpm00[n][m])/deltab);
    iFTm1b = kiFTpm00[n][m+1]
      + (fractionb * (kiFTpm00[n+1][m+1]-kiFTpm00[n][m+1])/deltab);
    iFT[1][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpm0m[n][m]
      + (fractionb * (krFTpm0m[n+1][m]-krFTpm0m[n][m])/deltab);
    rFTm1b = krFTpm0m[n][m+1]
      + (fractionb * (krFTpm0m[n+1][m+1]-krFTpm0m[n][m+1])/deltab);
    rFT[1][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpm0m[n][m]
      + (fractionb * (kiFTpm0m[n+1][m]-kiFTpm0m[n][m])/deltab);
    iFTm1b = kiFTpm0m[n][m+1]
      + (fractionb * (kiFTpm0m[n+1][m+1]-kiFTpm0m[n][m+1])/deltab);
    iFT[1][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpmmp[n][m]
      + (fractionb * (krFTpmmp[n+1][m]-krFTpmmp[n][m])/deltab);
    rFTm1b = krFTpmmp[n][m+1]
      + (fractionb * (krFTpmmp[n+1][m+1]-krFTpmmp[n][m+1])/deltab);
    rFT[0][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpmmp[n][m]
      + (fractionb * (kiFTpmmp[n+1][m]-kiFTpmmp[n][m])/deltab);
    iFTm1b = kiFTpmmp[n][m+1]
      + (fractionb * (kiFTpmmp[n+1][m+1]-kiFTpmmp[n][m+1])/deltab);
    iFT[0][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpmm0[n][m]
      + (fractionb * (krFTpmm0[n+1][m]-krFTpmm0[n][m])/deltab);
    rFTm1b = krFTpmm0[n][m+1]
      + (fractionb * (krFTpmm0[n+1][m+1]-krFTpmm0[n][m+1])/deltab);
    rFT[0][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpmm0[n][m]
      + (fractionb * (kiFTpmm0[n+1][m]-kiFTpmm0[n][m])/deltab);
    iFTm1b = kiFTpmm0[n][m+1]
      + (fractionb * (kiFTpmm0[n+1][m+1]-kiFTpmm0[n][m+1])/deltab);
    iFT[0][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTpmmm[n][m]
      + (fractionb * (krFTpmmm[n+1][m]-krFTpmmm[n][m])/deltab);
    rFTm1b = krFTpmmm[n][m+1]
      + (fractionb * (krFTpmmm[n+1][m+1]-krFTpmmm[n][m+1])/deltab);
    rFT[0][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTpmmm[n][m]
      + (fractionb * (kiFTpmmm[n+1][m]-kiFTpmmm[n][m])/deltab);
    iFTm1b = kiFTpmmm[n][m+1]
      + (fractionb * (kiFTpmmm[n+1][m+1]-kiFTpmmm[n][m+1])/deltab);
    iFT[0][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);
  } else if(initial==4) {
    rFTmb = krFTmppp[n][m]
      + (fractionb * (krFTmppp[n+1][m]-krFTmppp[n][m])/deltab);
    rFTm1b = krFTmppp[n][m+1]
      + (fractionb * (krFTmppp[n+1][m+1]-krFTmppp[n][m+1])/deltab);
    rFT[2][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmppp[n][m]
      + (fractionb * (kiFTmppp[n+1][m]-kiFTmppp[n][m])/deltab);
    iFTm1b = kiFTmppp[n][m+1]
      + (fractionb * (kiFTmppp[n+1][m+1]-kiFTmppp[n][m+1])/deltab);
    iFT[2][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmpp0[n][m]
      + (fractionb * (krFTmpp0[n+1][m]-krFTmpp0[n][m])/deltab);
    rFTm1b = krFTmpp0[n][m+1]
      + (fractionb * (krFTmpp0[n+1][m+1]-krFTmpp0[n][m+1])/deltab);
    rFT[2][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmpp0[n][m]
      + (fractionb * (kiFTmpp0[n+1][m]-kiFTmpp0[n][m])/deltab);
    iFTm1b = kiFTmpp0[n][m+1]
      + (fractionb * (kiFTmpp0[n+1][m+1]-kiFTmpp0[n][m+1])/deltab);
    iFT[2][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmppm[n][m]
      + (fractionb * (krFTmppm[n+1][m]-krFTmppm[n][m])/deltab);
    rFTm1b = krFTmppm[n][m+1]
      + (fractionb * (krFTmppm[n+1][m+1]-krFTmppm[n][m+1])/deltab);
    rFT[2][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmppm[n][m]
      + (fractionb * (kiFTmppm[n+1][m]-kiFTmppm[n][m])/deltab);
    iFTm1b = kiFTmppm[n][m+1]
      + (fractionb * (kiFTmppm[n+1][m+1]-kiFTmppm[n][m+1])/deltab);
    iFT[2][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmp0p[n][m]
      + (fractionb * (krFTmp0p[n+1][m]-krFTmp0p[n][m])/deltab);
    rFTm1b = krFTmp0p[n][m+1]
      + (fractionb * (krFTmp0p[n+1][m+1]-krFTmp0p[n][m+1])/deltab);
    rFT[1][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmp0p[n][m]
      + (fractionb * (kiFTmp0p[n+1][m]-kiFTmp0p[n][m])/deltab);
    iFTm1b = kiFTmp0p[n][m+1]
      + (fractionb * (kiFTmp0p[n+1][m+1]-kiFTmp0p[n][m+1])/deltab);
    iFT[1][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmp00[n][m]
      + (fractionb * (krFTmp00[n+1][m]-krFTmp00[n][m])/deltab);
    rFTm1b = krFTmp00[n][m+1]
      + (fractionb * (krFTmp00[n+1][m+1]-krFTmp00[n][m+1])/deltab);
    rFT[1][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmp00[n][m]
      + (fractionb * (kiFTmp00[n+1][m]-kiFTmp00[n][m])/deltab);
    iFTm1b = kiFTmp00[n][m+1]
      + (fractionb * (kiFTmp00[n+1][m+1]-kiFTmp00[n][m+1])/deltab);
    iFT[1][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmp0m[n][m]
      + (fractionb * (krFTmp0m[n+1][m]-krFTmp0m[n][m])/deltab);
    rFTm1b = krFTmp0m[n][m+1]
      + (fractionb * (krFTmp0m[n+1][m+1]-krFTmp0m[n][m+1])/deltab);
    rFT[1][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmp0m[n][m]
      + (fractionb * (kiFTmp0m[n+1][m]-kiFTmp0m[n][m])/deltab);
    iFTm1b = kiFTmp0m[n][m+1]
      + (fractionb * (kiFTmp0m[n+1][m+1]-kiFTmp0m[n][m+1])/deltab);
    iFT[1][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmpmp[n][m]
      + (fractionb * (krFTmpmp[n+1][m]-krFTmpmp[n][m])/deltab);
    rFTm1b = krFTmpmp[n][m+1]
      + (fractionb * (krFTmpmp[n+1][m+1]-krFTmpmp[n][m+1])/deltab);
    rFT[0][2] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmpmp[n][m]
      + (fractionb * (kiFTmpmp[n+1][m]-kiFTmpmp[n][m])/deltab);
    iFTm1b = kiFTmpmp[n][m+1]
      + (fractionb * (kiFTmpmp[n+1][m+1]-kiFTmpmp[n][m+1])/deltab);
    iFT[0][2] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmpm0[n][m]
      + (fractionb * (krFTmpm0[n+1][m]-krFTmpm0[n][m])/deltab);
    rFTm1b = krFTmpm0[n][m+1]
      + (fractionb * (krFTmpm0[n+1][m+1]-krFTmpm0[n][m+1])/deltab);
    rFT[0][1] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmpm0[n][m]
      + (fractionb * (kiFTmpm0[n+1][m]-kiFTmpm0[n][m])/deltab);
    iFTm1b = kiFTmpm0[n][m+1]
      + (fractionb * (kiFTmpm0[n+1][m+1]-kiFTmpm0[n][m+1])/deltab);
    iFT[0][1] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);

    rFTmb = krFTmpmm[n][m]
      + (fractionb * (krFTmpmm[n+1][m]-krFTmpmm[n][m])/deltab);
    rFTm1b = krFTmpmm[n][m+1]
      + (fractionb * (krFTmpmm[n+1][m+1]-krFTmpmm[n][m+1])/deltab);
    rFT[0][0] = rFTmb + (fractiont * (rFTm1b-rFTmb)/deltat);
    iFTmb = kiFTmpmm[n][m]
      + (fractionb * (kiFTmpmm[n+1][m]-kiFTmpmm[n][m])/deltab);
    iFTm1b = kiFTmpmm[n][m+1]
      + (fractionb * (kiFTmpmm[n+1][m+1]-kiFTmpmm[n][m+1])/deltab);
    iFT[0][0] = iFTmb + (fractiont * (iFTm1b-iFTmb)/deltat);
  }

  for(int i=0;i<3;i++){
    for(int j=0;j<3;j++){
      FT[i][j] = Complex_t(rFT[i][j], iFT[i][j]);
    }
  }


}
